/*
 * xhtmlbuffer.c
 * xhtmlite
 *
 * by Doug Wolfe
 * doug@wolfescience.com 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* TODO (#2#): image resize */

#include <gtk/gtk.h>

#include "xhtmlinterface.h"
#include "xhtmlbuffer.h"
#include "xhtmltag.h"
#include "support.h"
#include "base64.h"
#include "xhtmlundo.h"

#include <string.h>
#include <gdk-pixbuf/gdk-pixdata.h>

//for xml
#include <libxml/xmlreader.h>
#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>

#define MY_ENCODING "UTF-8"

static GList * undo_stack = NULL;
static GList * redo_stack = NULL;
static int image = -1;
GList *xhtml_images = NULL;

static GdkPixbuf * xhtmlite_image_repr_read_b64 (const gchar * uri_data)
{
	GdkPixbuf *pixbuf = NULL;
	GdkPixbufLoader *loader = NULL;
 
	gint j;
	gint k;
	gint l;
	gint b;
	gint len;
	gint eos = 0;
	gint failed = 0;
 
	guint32 bits;
 
	static const gchar B64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
	const gchar *btr = uri_data;
 
	gchar ud[4];
 
	guchar bd[57];
 
	loader = gdk_pixbuf_loader_new ();
 
	if (loader == NULL) return NULL;
 
	while (eos == 0) {
		l = 0;
		for (j = 0; j < 19; j++) {
			len = 0;
			for (k = 0; k < 4; k++) {
				while (isspace ((int) (*btr))) {
					if ((*btr) == '\0') break;
					btr++;
				}
				if (eos) {
					ud[k] = 0;
					continue;
				}
				if (((*btr) == '\0') || ((*btr) == '=')) {
					eos = 1;
					ud[k] = 0;
					continue;
				}
				ud[k] = 64;
				for (b = 0; b < 64; b++) { /* There must a faster way to do this... ?? */
					if (B64[b] == (*btr)) {
						ud[k] = (gchar) b;
						break;
					}
				}
				if (ud[k] == 64) { /* data corruption ?? */
					eos = 1;
					ud[k] = 0;
					continue;
				}
				btr++;
				len++;
			}
			bits = (guint32) ud[0];
			bits = (bits << 6) | (guint32) ud[1];
			bits = (bits << 6) | (guint32) ud[2];
			bits = (bits << 6) | (guint32) ud[3];
			bd[l++] = (guchar) ((bits & 0xff0000) >> 16);
			if (len > 2) {
				bd[l++] = (guchar) ((bits & 0xff00) >>  8);
			}
			if (len > 3) {
				bd[l++] = (guchar)  (bits & 0xff);
			}
		}
 
		if (!gdk_pixbuf_loader_write (loader, (const guchar *) bd, (size_t) l, NULL)) {
			failed = 1;
			break;
		}
	}
 
	gdk_pixbuf_loader_close (loader, NULL);
	if (!failed) pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
	return pixbuf;
}

static void xhtmlite_remove_old_color_tags(GtkTextBuffer *textbuffer, int color_type)
{
	/*removes old color tags from selection*/
	GtkTextIter iter, endselect, next_iter;
	gtk_text_buffer_get_selection_bounds(textbuffer, &iter, &endselect);	
	gtk_text_buffer_get_selection_bounds(textbuffer, &next_iter, &endselect);
	gtk_text_iter_forward_char(&next_iter);
	int n;

	while (!gtk_text_iter_equal(&iter, &endselect) && !gtk_text_iter_is_end(&iter))
	{
	GSList *tags = gtk_text_iter_get_tags(&iter);
	guint len = g_slist_length(tags);
		for(n = 0; n < len; n++)
		{
			GtkTextTag *tag = g_slist_nth_data (tags, n);
			if(g_strrstr(tag->name, "color")!= NULL)
			{
				if (g_strrstr(tag->name, "background-color")!= NULL)
				{
					if(color_type == COLOR_BACKGROUND)
					{
						printf("delete background-color:%s\n", tag->name);
						gtk_text_buffer_remove_tag(textbuffer, tag, &iter, &next_iter);
					}
				}
				else
				{
					if(color_type == COLOR_FOREGROUND)
					{
						printf("delete foreground-color:%s\n", tag->name);
						gtk_text_buffer_remove_tag(textbuffer, tag, &iter, &next_iter);
					}
				}
			}
		}
		gtk_text_iter_forward_char(&iter);
		gtk_text_iter_forward_char(&next_iter);
	}	
}

static void xhtmlite_remove_old_link_tags(GtkTextBuffer *textbuffer, GtkTextIter *start, GtkTextIter *end)
{
	/*removes old color tags from selection*/
	GtkTextIter iter = *gtk_text_iter_copy(start);
	GtkTextIter next_iter = *gtk_text_iter_copy(start);	
	GtkTextIter endselect = *gtk_text_iter_copy(end);
	gtk_text_iter_forward_char(&next_iter);
	int n;

	while (!gtk_text_iter_equal(&iter, &endselect) && !gtk_text_iter_is_end(&iter))
	{
	GSList *tags = gtk_text_iter_get_tags(&iter);
	guint len = g_slist_length(tags);
		for(n = 0; n < len; n++)
		{
			GtkTextTag *tag = g_slist_nth_data (tags, n);
			gchar *link = g_object_get_data(G_OBJECT(tag), "link");
			if(g_str_get_length(link) > 0)
			{
			gtk_text_buffer_remove_tag(textbuffer, tag, &iter, &next_iter);
			}
		}
		gtk_text_iter_forward_char(&iter);
		gtk_text_iter_forward_char(&next_iter);
	}	
}

static void xhtmlite_set_old_color(int color_type, GtkTextIter * iter, GtkColorSelection * color_selection)
{
	/*does not work*/
	GSList *tags = gtk_text_iter_get_tags(iter);
	GtkTextTag *tag = NULL;
	int n;
	guint len = g_slist_length(tags);
	GdkColor previous_color;
	for(n = 0; n < len; n++)
	{
		tag = g_slist_nth_data (tags, n);
		if(g_strrstr(tag->name, "color")!= NULL)
			{
				if (g_strrstr(tag->name, "background-color")!= NULL)
				{
					if(color_type == COLOR_BACKGROUND)
					{
						printf("set background-color:%s\n", tag->name);
						g_object_get (tag, "background_gdk", &previous_color, NULL);
						gtk_color_selection_set_previous_color(GTK_COLOR_SELECTION(color_selection), &previous_color);
					}
				}
				else
				{
					if(color_type == COLOR_FOREGROUND)
					{
						printf("set foreground-color:%s\n", tag->name);
						g_object_get (tag, "foreground_gdk", &previous_color, NULL);
						gtk_color_selection_set_previous_color(GTK_COLOR_SELECTION(color_selection), &previous_color);
					}
				}
			}
	}
}

static void xhtmlite_add_www_link(GtkTextBuffer *textbuffer,
					gchar *link, GtkTextIter *start, GtkTextIter *end)
{
	xhtmlite_remove_old_link_tags(textbuffer, start, end);
	if(g_str_get_length(link)>0)
	{
		gchar * tagname = g_strdup_printf("a href='%s'", link);
		printf("link:%s\n", link);
		/*g_strescape() g_strcompress()*/
		GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
		GtkTextTag * tag = gtk_text_tag_table_lookup(tagtable, tagname);
		/*look to see if tag exists and create it if not*/
		if(tag == NULL)
		{
			tag = gtk_text_tag_new (tagname);
			xhtmlite_setup_tag(tagtable,tag);
			GdkColor color;
			gdk_color_parse("blue", &color);
			g_object_set (tag, "foreground_gdk", &color, NULL);
			g_object_set (tag, "underline", PANGO_UNDERLINE_SINGLE,NULL);
			g_object_set_data (G_OBJECT (tag), "link", g_strdup(link));
		}
		if(tag != NULL)
		{
			gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
			xhtmlite_undo_add_undo(textbuffer, start, end, NULL, tag, XHTMLITE_UNDO_ADD_TAG);
		}
	}
}

static void xhtmlite_add_style_font_color(GtkTextBuffer *textbuffer,
				       gchar *color_string,
				       const GtkTextIter *start,
				       const GtkTextIter *end)
{
	GtkTextTag *tag;
	GdkColor color;
	color_string = g_utf8_strdown(color_string, -1);
	if(gdk_color_parse(color_string, &color))
	{
			//make string and convert color to hexadecimal
		color_string = g_strdup_printf("span color='%s'", color_string);
		printf("color_string:%s\n", color_string);
		GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
		tag = gtk_text_tag_table_lookup(tagtable, color_string);
						//look to see if tag exists and create it if not
		if(tag == NULL)
		{
			tag = gtk_text_tag_new (color_string);
			xhtmlite_setup_tag(tagtable,tag);
			g_object_set (tag, "foreground_gdk", &color, NULL);
		}
		if(tag != NULL)
		{
			gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
		}
	}
	else
	{
		printf("color did not parse correctly:%s\n", color_string);
	}
}

static void xhtmlite_add_style_font_background_color(GtkTextBuffer *textbuffer,
					gchar *color_string,
				       const GtkTextIter *start,
				       const GtkTextIter *end)
{
	GtkTextTag *tag;
	GdkColor color;
	color_string = g_utf8_strdown(color_string, -1);
	if(gdk_color_parse(color_string, &color))
	{
		//make string and convert color to hexadecimal
		color_string = g_strdup_printf("span background-color='%s'", color_string);
		printf("color_string:%s\n", color_string);

		GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
		tag = gtk_text_tag_table_lookup(tagtable, color_string);
						//look to see if tag exists and create it if not
		if(tag == NULL)
		{
			tag = gtk_text_tag_new (color_string);
			xhtmlite_setup_tag(tagtable,tag);
			g_object_set (tag, "background_gdk", &color, NULL);
		}
		if(tag != NULL)
		{
			gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
		}
	}
	else
	{
		printf("color did not parse correctly:%s\n", color_string);
	}
}

static void xhtmlite_add_style_font_face(GtkTextBuffer *textbuffer,
				      const gchar *face,
				       const GtkTextIter *start,
				       const GtkTextIter *end)
{
	GtkTextTag *tag;
	//create the font-size tag name
	//do not lowercase this attribute as font face is case sensitive. For instance: "Sans"
	gchar *fontfamily = g_strdup_printf("span face='%s'", face);
	printf("fontfamily:%s\n", fontfamily);
	//look to see if tag exists and create if not
	GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
	tag = gtk_text_tag_table_lookup(tagtable, fontfamily);
	if(tag == NULL)
	{
		printf("fontfamily tag not found: %s\n", fontfamily);
		tag = gtk_text_tag_new (fontfamily);
		xhtmlite_setup_tag(tagtable,tag);
		g_object_set (tag, "family", attribute, NULL);
	}
	if(tag != NULL)
	{
		gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
	}
}

static void xhtmlite_add_style_font_size(GtkTextBuffer *textbuffer,
				      const gchar *font_size,
				      const GtkTextIter *start,
				      const GtkTextIter *end)
{
	GtkTextTag *tag;
	font_size = g_utf8_strdown(font_size, -1);
	//create the font-size tag name
	gchar *fontsize_string =g_strdup_printf("span font-size='%s'", font_size);
	printf("fontsize_string:%s\n", fontsize_string);
					//look to see if tag exists and release error if it is not found
	GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
	tag = gtk_text_tag_table_lookup(tagtable, fontsize_string);
	if(tag == NULL)
	{
		printf("font-size tag not found: %s\n", fontsize_string);
	}
	if(tag != NULL)
	{
		gtk_text_buffer_apply_tag(textbuffer, tag, start, end);
	}
}

static void xhtmlite_process_font_tag(GtkTextBuffer * textbuffer, xmlTextReaderPtr reader,
			     const GtkTextIter *start, const GtkTextIter *end)
{
	gchar *attribute;
	attribute = (gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"background-color");
	if(attribute != NULL)
	{
		xhtmlite_add_style_font_background_color(textbuffer, attribute, start, end);
	}
	attribute = (gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"color");
	if(attribute != NULL)
	{
		xhtmlite_add_style_font_color(textbuffer, attribute, start, end);
	}
	attribute = (gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"font-size");
	if(attribute != NULL)
	{
		xhtmlite_add_style_font_size(textbuffer, attribute, start, end);
	}
	attribute = (gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"face");
	if(attribute != NULL)
	{
		xhtmlite_add_style_font_face(textbuffer, attribute, start, end);
	}
}

static void xhtmlite_parse_and_process_style_attribute(GtkTextBuffer * textbuffer, xmlTextReaderPtr reader,
			     const GtkTextIter *start, const GtkTextIter *end)
{
	GtkTextTag *tag;
	const gchar *attribute;
	attribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"style");
	if(attribute != NULL)
	{
		gchar *font_size;
		gchar *font_face;
		gchar *color;
		gchar *background_color;

		gchar **split_style = g_strsplit_set (attribute, ";",-1);
		int n;
		for (n = 0; split_style[n] ; n++)
		{
			printf("split_style(n:%d):%s\n", n, split_style[n]);
			gchar **split_substyle = g_strsplit_set (split_style[n], ":",-1);
			printf("split_style(%d):%s\n", 0, split_substyle[0]);
			printf("split_style(%d):%s\n", 1, split_substyle[1]);

			if(strcmp(split_substyle[0], "background-color") == 0)
			{
				xhtmlite_add_style_font_background_color(textbuffer, split_substyle[1], start, end);
			}
			else if(strcmp(split_substyle[0], "color") == 0)
			{
				xhtmlite_add_style_font_color(textbuffer, split_substyle[1], start, end);
			}
			else if(strcmp(split_substyle[0], "font-size") == 0)
			{
				xhtmlite_add_style_font_size(textbuffer, split_substyle[1], start, end);
			}
			else if(strcmp(split_substyle[0], "face") == 0)
			{
				xhtmlite_add_style_font_face(textbuffer, split_substyle[1], start, end);
			}
		}
	}
}

static void xhtmlite_process_a_link_tag(GtkTextBuffer * textbuffer, xmlTextReaderPtr reader,
			     GtkTextIter *start, GtkTextIter *end)
{
	/* TODO (#1#): work on the reading of this tag */
	GtkTextTag *tag;
	gchar *attribute;
	attribute = (gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"href");
	if(attribute != NULL)
	{
		xhtmlite_add_www_link(textbuffer, attribute, start, end);
	}
}

static gboolean xhtmlite_tag_ends_here (GtkTextTag *tag, GtkTextIter *iter, GtkTextIter *next_iter)
{
	return (gtk_text_iter_has_tag (iter, tag) && !gtk_text_iter_has_tag (next_iter, tag) || gtk_text_iter_is_end(next_iter));
}

static gboolean xhtmlite_toggle_selected_tag (GtkTextBuffer *textbuffer, gchar *tag_name)
{
	GtkTextIter start, end;
	gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end);
	GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
	GtkTextTag *tag = gtk_text_tag_table_lookup(tagtable, tag_name);

	while(!gtk_text_iter_equal(&start, &end))
	{
		if(!gtk_text_iter_has_tag (&start, tag))
		{
			gtk_text_buffer_apply_tag (textbuffer, tag, &start, &end);
			xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, tag, XHTMLITE_UNDO_ADD_TAG);
			return (TRUE);
		}
		gtk_text_iter_forward_char(&start);
	}
	gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end);
	gtk_text_buffer_remove_tag (textbuffer, tag, &start, &end);
	xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, tag, XHTMLITE_UNDO_REMOVE_TAG);
	return (FALSE);
}

static gboolean xhtmlite_set_selected_tag (GtkTextBuffer *textbuffer, gchar * tag_name, gboolean enable)
{
	GtkTextIter start, end;
	gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end);
	GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
	GtkTextTag *tag = gtk_text_tag_table_lookup(tagtable, tag_name);

	if(enable)
	{
		gtk_text_buffer_apply_tag (textbuffer, tag, &start, &end);
		xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, tag, XHTMLITE_UNDO_ADD_TAG);
		return (TRUE);
	}

	gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end);
	gtk_text_buffer_remove_tag (textbuffer, tag, &start, &end);
	xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, tag, XHTMLITE_UNDO_REMOVE_TAG);
	return (FALSE);
}

static gchar* xhtmlite_append_style (gchar *style_string, gchar *style_name, gchar *style_value)
{
	if(style_string==NULL)
	{
		return g_strdup_printf("%s:%s", style_name, style_value);
	}
	else
	{
		return g_strdup_printf("%s;%s:%s", style_string, style_name, style_value);
	}
}


xmlChar *xhtmlite_convert_xml_input(const char *in, const char *encoding);

void xhtmlite_write_start_element(xmlTextWriterPtr writer, GtkTextTag *tag)
{
	//write start xml element
	int rc;
	if(g_strrstr(tag->name, "span\0")!= NULL)
	{
		//set font with attributes
		rc = xmlTextWriterStartElement(writer, BAD_CAST "span");
		if (rc < 0) {
			printf	("xmlTextWriterStartElement: Error writing 'span' element\n");
			return;
		}
		
		gchar *style = NULL;
		gchar *value;
		if(g_strrstr(tag->name, "face\0")!= NULL)
		{
			g_object_get (tag, "family", &value, NULL);
			style = xhtmlite_append_style (style, "face", value);
		}
		
		if(g_strrstr(tag->name, "font-size\0")!= NULL)
		{
			value = g_object_get_data(G_OBJECT(tag), "font-size_absolute");
			style = xhtmlite_append_style (style, "font-size", value);
		}
	
		if(g_strrstr(tag->name, "color")!= NULL)
		{
		//make string and convert color to hexadecimal
			GdkColor *color;
			if (g_strrstr(tag->name, "background-color")!= NULL)
			{
				g_object_get (tag, "background_gdk", &color, NULL);
				value = g_strdup_printf("#%x%x%x", color->red/256, color->green/256, color->blue/256);
				style = xhtmlite_append_style (style, "background-color", value);
			}
			else 
			{
				g_object_get (tag, "foreground_gdk", &color, NULL);
				value = g_strdup_printf("#%x%x%x", color->red/256, color->green/256, color->blue/256);
				style = xhtmlite_append_style (style, "color", value);
			}
		}

		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "style",
				BAD_CAST style);
		if (rc < 0) {
			printf("xmlTextWriterWriteAttribute: Error writing 'face' attribute\n");
			return;
		}
	}
	else if(g_strrstr(tag->name, "font\0")!= NULL)
	{
		//set font with attributes
		rc = xmlTextWriterStartElement(writer, BAD_CAST "font");
		if (rc < 0) {
			printf	("xmlTextWriterStartElement: Error writing 'font' element\n");
			return;
		}

		if(g_strrstr(tag->name, "face\0")!= NULL)
		{
			gchar *fontfamily;
			g_object_get (tag, "family", &fontfamily, NULL);
			rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "face",
					BAD_CAST fontfamily);
			if (rc < 0) {
				printf("xmlTextWriterWriteAttribute: Error writing 'face' attribute\n");
				return;
			}
		}
		
		if(g_strrstr(tag->name, "font-size\0")!= NULL)
		{
			gchar *fontsize = g_object_get_data(G_OBJECT(tag), "font-size_absolute");
			rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "font-size",
					BAD_CAST fontsize);
			if (rc < 0) {
				printf("xmlTextWriterWriteAttribute: Error writing absolute 'font-size'\n");
				return;
			}
		}
	
		if(g_strrstr(tag->name, "color")!= NULL)
		{
		//make string and convert color to hexadecimal
			GdkColor *color;
			gchar *color_string;
			if (g_strrstr(tag->name, "background-color")!= NULL)
			{
				g_object_get (tag, "background_gdk", &color, NULL);
				color_string = g_strdup_printf("#%x%x%x", color->red/256, color->green/256, color->blue/256);
				printf("background-color=%s\n", color_string);
				/* Add an attribute with name "version" and value "1.0" to ORDER. */
				rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "background-color",
						BAD_CAST color_string);
				if (rc < 0) {
					printf("xmlTextWriterWriteAttribute: Error writing 'background-color' tag\n");
					return;
				}
			}
			else
			{
				g_object_get (tag, "foreground_gdk", &color, NULL);
				color_string = g_strdup_printf("#%x%x%x", color->red/256, color->green/256, color->blue/256);
				printf("color=%s\n", color_string);
				/* Add an attribute with name "version" and value "1.0" to ORDER. */
				rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "color",
						BAD_CAST color_string);
				if (rc < 0) {
					printf("xmlTextWriterWriteAttribute: Error writing 'color' tag\n");
					return;
				}
			}
		}
	}
	else if(g_strrstr(tag->name, "a href")!= NULL)
	{
		//set font with attributes
		rc = xmlTextWriterStartElement(writer, BAD_CAST "a");
		if (rc < 0) {
			printf	("xmlTextWriterStartElement: Error writing 'a href' element\n");
			return;
		}
		gchar *link = g_object_get_data(G_OBJECT(tag), "link");
		if(strcmp(tag->name, "link") != 0)
		{
			rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "href",
					BAD_CAST link);
			if (rc < 0) {
				printf("xmlTextWriterWriteAttribute: Error writing 'href' attribute\n");
				return;
			}
		}
	}
	else
	{
		rc = xmlTextWriterStartElement(writer, BAD_CAST tag->name);
		if (rc < 0) {
			printf	("xmlTextWriterStartElement: Error at writing tag '%s'\n", tag->name);
			return;
		}
	}
}

void
xhtmlite_write_end_element(xmlTextWriterPtr writer, GtkTextTag *tag)
{
	/* Close the element */
	int rc;
	rc = xmlTextWriterEndElement(writer);
	if (rc < 0) {
		printf("testXmlwriterFilename: Error at xmlTextWriterEndElement\n");
		return;
	}
}

/**
* ConvertInput:
 * @in: string in a given encoding
 * @encoding: the encoding used
 *
 * Converts @in into UTF-8 for processing with libxml2 APIs
 *
 * Returns the converted UTF-8 string, or NULL in case of error.
 */
xmlChar* 
xhtmlite_convert_xml_input(const char *in, const char *encoding)
{
	xmlChar *out;
	int ret;
	int size;
	int out_size;
	int temp;
	xmlCharEncodingHandlerPtr handler;

	if (in == 0)
		return 0;

	handler = xmlFindCharEncodingHandler(encoding);

	if (!handler) {
		printf("ConvertInput: no encoding handler found for '%s'\n",
		       encoding ? encoding : "");
		return 0;
	}

	size = (int) g_str_get_length(in) + 1;
	out_size = size * 2 - 1;
	out = (unsigned char *) xmlMalloc((size_t) out_size);

	if (out != 0) {
		temp = size - 1;
		ret = handler->input(out, &out_size, (const xmlChar *) in, &temp);
		if ((ret < 0) || (temp - size + 1)) {
			if (ret < 0) {
				printf("ConvertInput: conversion wasn't successful.\n");
			} else {
				printf
						("ConvertInput: conversion wasn't successful. converted: %i octets.\n",
						temp);
			}

			xmlFree(out);
			out = 0;
		}
		else {
			out = (unsigned char *) xmlRealloc(out, out_size + 1);
			out[out_size] = 0;  /*null terminating out */
		}
	} else {
		printf("ConvertInput: no mem\n");
	}

	return out;
}


int *xhtmlite_add_image_to_stack(GList *list, GtkImage *image)
{
	
}

gchar *xhtmlite_write_range (GtkTextBuffer *textbuffer, GtkTextIter *start, GtkTextIter *end, gboolean css)
{
	int rc;
	xmlTextWriterPtr writer;
	xmlBufferPtr buf;
	xmlChar *tmp;
	printf("Started write of xml.\n");
    /* Create a new XML buffer, to which the XML document will be
    * written */
	buf = xmlBufferCreate();
	if (buf == NULL) {
		printf("testXmlwriterMemory: Error creating the xml buffer\n");
		return;
	}

    /* Create a new XmlWriter for memory, with no compression.
    * Remark: there is no compression for this kind of xmlTextWriter */
	writer = xmlNewTextWriterMemory(buf, 0);
	if (writer == NULL) {
		printf("testXmlwriterMemory: Error creating the xml writer\n");
		return;
	}

	/* Start the document with the xml default for the version*/
	rc = xmlTextWriterStartDocument(writer, NULL, MY_ENCODING, NULL);
	if (rc < 0) {
		printf("testXmlwriterMemory: Error at xmlTextWriterStartDocument\n");
		return;
	}

	/* Start an element named "HTML". Since thist is the first
	* element, this will be the root element of the document. */
	printf("write start document element\n");
	rc = xmlTextWriterStartElement(writer, BAD_CAST "html");
	if (rc < 0) {
		printf("testXmlwriterMemory: Error at xmlStartElement\n");
		return;
	}
	
	GList *tag_stack = NULL;
	GList *replay_stack = NULL;
	
	int start_offset = gtk_text_iter_get_offset (start);
	int end_offset = gtk_text_iter_get_offset (end);
	
	GtkTextIter iter, next_iter;
	gtk_text_buffer_get_iter_at_offset (textbuffer, &iter, start_offset);
	gtk_text_buffer_get_iter_at_offset (textbuffer, &next_iter, start_offset);
	gtk_text_iter_forward_char(&next_iter);

	GSList *tags =  gtk_text_iter_get_tags(start);
	int n;
	guint len =  g_slist_length(tags);
	printf ("start tags len:%d\n", len);
	for(n = 0; n < len; n++)
	{
		GtkTextTag *start_tag = g_slist_nth_data (tags, n);
		//if (TVTagTable.TagIsIgnored (start_tag))
 		//	continue;
  
		if (!gtk_text_iter_toggles_tag(start, start_tag))
		{
 			//tag_stack.Push (start_tag);
			tag_stack=g_list_prepend (tag_stack, start_tag); //push iter onto the beginning of the list
			xhtmlite_write_start_element(writer, start_tag);
			//xml.xhtmlite_write_start_element (TagXmlStart(start_tag)); 
			
 			//TagXmlAttribute(start_tag, xml);
		}
	}

	while (!gtk_text_iter_equal(&iter, end) && !gtk_text_iter_is_end(&iter))
	{
		tags =  gtk_text_iter_get_tags(&iter);
		len =  g_slist_length(tags);
		for(n = 0; n < len; n++)
		{
			GtkTextTag *tag = g_slist_nth_data (tags, n);

				//if (TVTagTable.TagIsIgnored (start_tag))
				//	continue;
			if (gtk_text_iter_begins_tag(&iter, tag))
			{
				//tag_stack.Push (start_tag);
				tag_stack = g_list_prepend (tag_stack, tag); //push iter on to the beginning of the list
				printf("tag added\n");
				xhtmlite_write_start_element(writer, tag);
				//TagXmlAttribute(start_tag, xml);
			}
		}
 		// Hidden character representing an anchor
		gunichar iter_char = gtk_text_iter_get_char(&iter);
		if (iter_char ==  0xFFFC) //for a not valid unicode characters, such as an image
		{
 		
			GdkPixbuf *mypixbuf = gtk_text_iter_get_pixbuf (&iter);
		
			//get file_string
			if(mypixbuf != NULL)
			{
				printf("write start image element\n");
				rc = xmlTextWriterStartElement(writer, BAD_CAST "img");
				if (rc < 0) {
					printf("testXmlwriterMemory: Error at xmlStartElement: img\n");
					return;
				}
				xhtml_image *image= (xhtml_image*)g_object_get_data(G_OBJECT(mypixbuf), "xhtml_image");
				if(css)
				{
					rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "class", (const  xmlChar *)image->name);
				}
				else
				{
				printf("not css! This will try to write the image directly.\n");
					//printf("file_string:%s\n", file_string);
					//xmlTextWriterWriteFormatAttribute does not seem to work here
					rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "class", (const  xmlChar *)image->name);
					rc = xmlTextWriterStartAttribute(writer, BAD_CAST "src");
					if (rc < 0) {
						//printf("testXmlwriterFilename: Error at xmlTextWriterWriteAttribute\n");
						return;
					}
					rc = xmlTextWriterWriteString(writer, BAD_CAST "data:image;base64,");
					if (rc < 0) {
						//printf("testXmlwriterFilename: Error at xmlTextWriterWriteAttribute\n");
						return;
					}
					rc = xmlTextWriterWriteString(writer, BAD_CAST image->base64);
					if (rc < 0) {
						//printf("testXmlwriterFilename: Error at xmlTextWriterWriteAttribute\n");
						return;
					}
					rc = xmlTextWriterEndAttribute	(writer);
					if (rc < 0) {
						//printf("testXmlwriterFilename: Error at xmlTextWriterWriteAttribute\n");
						return;
					}
				}
				rc = xmlTextWriterEndElement(writer);
				if (rc < 0) {
					printf("testXmlwriterFilename: Error at xmlTextWriterEndElement:img\n");
					return;
				}
			}
			// if (iter.ChildAnchor != null)
			// {
			// 	//Console.WriteLine(iter.ChildAnchor.ToString());
			// 	string serialize = (string) iter.ChildAnchor.Data["serialize"];
			// 	if (serialize != null)
			// 		xml.WriteRaw (serialize);
			// }
		}
		else
		{
			//tmp = ConvertInput(BAD_CAST &iter_char, MY_ENCODING);
			static gchar charbuf[10];
			int charlen;
			charlen = g_unichar_to_utf8 (iter_char, charbuf);
			charbuf[charlen] = '\0';
			xmlTextWriterWriteString(writer, BAD_CAST charbuf);
		}

		for(n = 0; n < len; n++)
		{
			GtkTextTag *tag = g_slist_nth_data (tags, n);

			//if (TVTagTable.TagIsIgnored (tag))
			//	continue;
			if (xhtmlite_tag_ends_here (tag, &iter, &next_iter))
			{
				// Unwind until the ended tag.
				// Put any intermediate tags
				// into a replay queue...
				printf ("tag_stack_len:%d\n", g_list_length(tag_stack));
				while (g_list_length(tag_stack) > 0)
				{
					GtkTextTag *existing_tag = g_list_nth_data(tag_stack, 0);
					tag_stack=g_list_remove(tag_stack, existing_tag);
					if(!xhtmlite_tag_ends_here (existing_tag, &iter, &next_iter))
					{
						replay_stack=g_list_prepend (replay_stack, existing_tag); //push iter on to the list the beginning of the list
					}
					/* Close the element */
					xhtmlite_write_end_element(writer, tag);
					printf ("tag_stack_len:%d\n", g_list_length(tag_stack));
				}
					// Replay the replay queue.
					// Restart any tags that
					// overlapped with the ended
					// tag...
				while (g_list_length(replay_stack) > 0)
				{
					GtkTextTag *replay_tag = g_list_nth_data(replay_stack, 0);
					replay_stack= g_list_remove(replay_stack, replay_tag);
					tag_stack=g_list_prepend (tag_stack, replay_tag); //push iter on to the list the beginning of the list
					xhtmlite_write_start_element(writer, replay_tag);
					//TagXmlAttribute(replay_tag, xml);
				}
			}
		}


		if (gtk_text_iter_ends_line(&iter))
		{
			/* Start an element named "br" for linebreak. */
			rc = xmlTextWriterStartElement(writer, BAD_CAST "br");
			if (rc < 0) {
				printf	("testXmlwriterFilename: Error at xmlTextWriterStartElement\n");
				return;
			}
			/* Close the element named "br" for linebreak. */
			rc = xmlTextWriterEndElement(writer);
			if (rc < 0) {
				printf("testXmlwriterFilename: Error at xmlTextWriterEndElement\n");
				return;
			}
		}
		gtk_text_iter_forward_char(&iter);
		gtk_text_iter_forward_char(&next_iter);
	}

	/* Here we could close the elements and since we do not want to
	* write any other elements, we simply call xmlTextWriterEndDocument,
	* which will do all the work. */
	rc = xmlTextWriterEndDocument(writer);
	if (rc < 0) {
		printf("testXmlwriterMemory: Error at xmlTextWriterEndDocument\n");
		return;
	}
	
	xmlFreeTextWriter(writer);

	gchar *charbuffer= g_strdup((const gchar *)buf->content);
	xmlBufferFree(buf);	
	return charbuffer;
}

gchar *xhtmlite_write (GtkTextBuffer *textbuffer, gboolean css)
{
	GtkTextIter start, end;
	gtk_text_buffer_get_start_iter (textbuffer, &start);
	gtk_text_buffer_get_end_iter (textbuffer, &end);

	return xhtmlite_write_range (textbuffer, &start, &end, css);
}

gchar *xhtmlite_write_as_markup (GtkTextBuffer *textbuffer, gboolean css)
{
	gchar *xhtmlite = xhtmlite_write(textbuffer, css);
	strrepl(xhtmlite, "<html>", "", xhtmlite);
	strrepl(xhtmlite, "</html>\n", "", xhtmlite);
	strrepl(xhtmlite, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", "", xhtmlite);
	return xhtmlite;
}

static GtkTextTagTable *tagtable_static = NULL;
GtkTextBuffer* 
xhtmlite_create_buffer (GtkTextView *textview, gboolean writeable)
{
	//create and set buffer
	//add if for tag table already created
	if (tagtable_static==NULL) 
    {
      tagtable_static =  gtk_text_tag_table_new();
	   xhtmlite_fill_tagtable(tagtable_static);
    }
	GtkTextBuffer *textbuffer = gtk_text_buffer_new (tagtable_static);
	g_object_set_data(G_OBJECT(textbuffer), "name", "xhtmlite_buffer");
	if(textview != NULL)
		gtk_text_view_set_buffer(textview, textbuffer);

	if(writeable)
	{
		xhtmlite_undo_initialize_undos(textbuffer);
	}
	//gtk_text_buffer_set_modified (textbuffer, FALSE);
	return textbuffer;
}

void xhtmlite_toggle_bold (GtkTextBuffer *textbuffer, gboolean enable)
{
	xhtmlite_set_selected_tag (textbuffer, "b", enable);
}

void xhtmlite_toggle_italic (GtkTextBuffer *textbuffer, gboolean enable)
{
	xhtmlite_set_selected_tag (textbuffer, "i", enable);
}

void xhtmlite_toggle_underline (GtkTextBuffer *textbuffer, gboolean enable)
{
	xhtmlite_set_selected_tag (textbuffer, "u", enable);
}

void xhtmlite_toggle_strikethrough (GtkTextBuffer *textbuffer, gboolean enable)
{
	xhtmlite_set_selected_tag (textbuffer, "del", enable); //Strikethrough
}

void xhtmlite_toggle_superscript (GtkTextBuffer *textbuffer, gboolean enable)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "sub", &beginselect, &endselect);
	xhtmlite_set_selected_tag(textbuffer, "sup", enable);
}

void xhtmlite_toggle_subscript (GtkTextBuffer *textbuffer, gboolean enable)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "sup", &beginselect, &endselect);
	xhtmlite_set_selected_tag(textbuffer, "sub", enable);
}

void xhtmlite_adjust_fontsize_absolute(GtkTextBuffer *textbuffer, gchar *fontsize)
{
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='xx-small'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='x-small'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='small'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='medium'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='large'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='x-large'", &beginselect, &endselect);
	gtk_text_buffer_remove_tag_by_name(textbuffer, "span font-size='xx-large'", &beginselect, &endselect);
	
	xhtmlite_toggle_selected_tag(textbuffer, g_strdup_printf("span font-size='%s'", fontsize));
}

void xhtmlite_change_fontfamily (GtkTextBuffer *textbuffer, GtkWidget *textview, gchar *fontfamily)
{
	GtkTextIter start, end;
	gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end);
	gchar *fontfamily_tagname;
	GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
	GtkTextTag *tag;
	
	//remove all old fontfamily tags
	PangoContext *textview_context = gtk_widget_get_pango_context(textview);
	PangoFontFamily **families;
	int families_len, n;
	pango_context_list_families (textview_context, &families, &families_len);
	for (n = 0;  n < families_len;  n++)
	{
		const gchar *fontname =  pango_font_family_get_name (families[n]);
		fontfamily_tagname = g_strdup_printf("span face='%s'", fontname);
		tag = gtk_text_tag_table_lookup(tagtable, fontfamily_tagname);
		if (tag != NULL)
		{	
			gtk_text_buffer_remove_tag(textbuffer, tag, &start, &end);
			xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, tag, XHTMLITE_UNDO_REMOVE_TAG);
		}
	}

	//new tag
	fontfamily_tagname = g_strdup_printf("span face='%s'", fontfamily);
	tag = gtk_text_tag_table_lookup(tagtable, fontfamily_tagname);
	//look to see if tag exists and if doesn't exist, create it 
	if(tag == NULL)
	{
		tag = gtk_text_tag_new(fontfamily_tagname);
		xhtmlite_setup_tag(tagtable,tag);
		g_object_set (tag, "family", fontfamily, NULL);
	}
	gtk_text_buffer_apply_tag (textbuffer, tag, &start, &end);
	xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, tag, XHTMLITE_UNDO_ADD_TAG);
}

static GtkWidget *xhtmlite_colorselectiondialog = NULL;

void xhtmlite_open_font_color_dialog (GtkWindow *parent_window, GtkTextBuffer *textbuffer)
{
	if(xhtmlite_colorselectiondialog == NULL)
	{
		xhtmlite_colorselectiondialog = create_xhtmlite_colorselectiondialog();
		gtk_window_set_transient_for(GTK_WINDOW(xhtmlite_colorselectiondialog), parent_window);
	}

	GtkWidget *color_selection = lookup_widget(GTK_WIDGET(xhtmlite_colorselectiondialog), "xhtmlite_color_selection");

	GtkTextIter start, end;
	gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end);	
	
	/*xhtmlite_set_old_color(COLOR_FOREGROUND, &beginselect, GTK_COLOR_SELECTION(color_selection));*/
	
	if (gtk_dialog_run (GTK_DIALOG (xhtmlite_colorselectiondialog)) == GTK_RESPONSE_OK)
	{
		xhtmlite_remove_old_color_tags(textbuffer, COLOR_FOREGROUND);
		
		GdkColor color;
		gtk_color_selection_get_current_color(GTK_COLOR_SELECTION(color_selection), &color);

		//make string and convert color to hexadecimal
		gchar  *color_string = g_strdup_printf("span color='#%x%x%x'", color.red/256, color.green/256, color.blue/256);
		printf("%s\n", color_string);

		GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
		GtkTextTag *tag = gtk_text_tag_table_lookup(tagtable, (gchar *)color_string);
		
		//look to see if tag exists and create it if not
		if(tag == NULL)
		{
			tag = gtk_text_tag_new (color_string);
			xhtmlite_setup_tag(tagtable,tag);
			g_object_set (tag, "foreground_gdk", &color, NULL);
		}
	
		gtk_text_buffer_apply_tag (textbuffer, tag, &start, &end);
		xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, tag, XHTMLITE_UNDO_ADD_TAG);
	}
	gtk_widget_hide(xhtmlite_colorselectiondialog);
}

void xhtmlite_open_highlight_color_dialog  (GtkWindow *parent_window, GtkTextBuffer *textbuffer)
{
	if(xhtmlite_colorselectiondialog == NULL)
	{
		xhtmlite_colorselectiondialog = create_xhtmlite_colorselectiondialog();
		gtk_window_set_transient_for(GTK_WINDOW(xhtmlite_colorselectiondialog), parent_window);
	}

	GtkWidget *color_selection = lookup_widget(GTK_WIDGET(xhtmlite_colorselectiondialog), "xhtmlite_color_selection");

	GtkTextIter start, end;
	gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end);	
	
	/*xhtmlite_set_old_color(COLOR_BACKGROUND, &beginselect, GTK_COLOR_SELECTION(color_selection));*/

	if (gtk_dialog_run (GTK_DIALOG (xhtmlite_colorselectiondialog)) == GTK_RESPONSE_OK)
	{
		xhtmlite_remove_old_color_tags(textbuffer, COLOR_BACKGROUND);
		
		GdkColor color;
		gtk_color_selection_get_current_color(GTK_COLOR_SELECTION(color_selection), &color);
    
		//make string and convert color to hexadecimal
		gchar  *color_string = g_strdup_printf("span background-color='#%x%x%x'", color.red/256, color.green/256, color.blue/256);
		printf("%s\n", color_string);

		GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
		GtkTextTag *tag = gtk_text_tag_table_lookup(tagtable, (gchar *)color_string);
		
		//look to see if tag exists and create it if not
		if(tag == NULL)
		{
			tag = gtk_text_tag_new (color_string);
			xhtmlite_setup_tag(tagtable,tag);
			g_object_set (tag,
				      "background_gdk", &color,
				      NULL);
		}
	
		gtk_text_buffer_apply_tag (textbuffer, tag, &start, &end);
		xhtmlite_undo_add_undo(textbuffer, &start, &end, NULL, tag, XHTMLITE_UNDO_ADD_TAG);		
	}
	gtk_widget_hide(xhtmlite_colorselectiondialog);
}

xhtml_image*
xhtmlite_pixbuf_add_to_memory(gchar *filename, gchar *name, gchar *base64)
{
	xhtml_image *image_data;
	printf("found \n");
	if (xhtml_images != NULL)
	{
		int n;
		for (n=0;n<g_list_length(xhtml_images); n++)
		{
			image_data = g_list_nth_data(xhtml_images, n);
			printf("filename:%s", image_data->filename);
			if(g_strrstr(name, image_data->name) != NULL)
			{
				printf("found \n");
				return image_data;		
			}
		}
	}
	
	if(base64 == NULL)
		return NULL;
	
	image_data = malloc(sizeof(xhtml_image));
	image_data->name = g_strdup(filename); 
	image_data->filename = g_strdup(filename);
	image_data->base64 = g_strdup(base64);
	
	printf("filename:%s\n", image_data->filename);
	xhtml_images =  g_list_append(xhtml_images, image_data);
	return image_data;
}

static GdkPixbuf*
xhtmlite_file_decode_pixbuf_from_xhtml (const gchar *src, const gchar *class)
{
	//TODO: FIX THIS!!!
	//reads picture in base64 format or already stored.
	//Does NOT read picture stored on drive or remotely yet.
	GdkPixbuf *mypixbuf = NULL;
	gchar *filename;
	gchar *name;
	gchar *base64;
	xhtml_image *image_data;

	printf("testing read file. . .");
	if(src != NULL)
	{
		//gchar image_encoded[g_str_get_length(src)];
		gchar *base64find = g_strstr_len (src, g_str_get_length(src), "base64,");
		//if base64 is not found, the src is most likely a uri
		strrepl(base64find, "base64,", "", base64);
		gchar *file_string = g_strdup(base64);
		mypixbuf = xhtmlite_image_repr_read_b64 (file_string);
	
		xhtml_image *image_data = xhtmlite_pixbuf_add_to_memory((gchar *)class, (gchar *)class, base64);


		g_object_set_data (G_OBJECT (mypixbuf), "xhtml_image", image_data); 
	}
	else if(class != NULL)
	{
		printf("class\n");
		xhtml_image *image_data = xhtmlite_pixbuf_add_to_memory((gchar *)class, (gchar *)class, NULL);
		
		if (image_data == NULL)
			return NULL;
		mypixbuf = xhtmlite_image_repr_read_b64 (image_data->base64);
		g_object_set_data (G_OBJECT (mypixbuf), "xhtml_image", image_data);
	}

	return mypixbuf;
}

GList *import_tag_stack = NULL;

int offset;
int xhtmlite_insert (GtkTextBuffer *textbuffer, gchar *xml, GtkTextIter *start, gboolean markup)
{
	//Fix White Spaces
	strrepl(xml, "\r", "", xml);
	strrepl(xml, "\n", "", xml);
	
	strrepl(xml, "<br />", "\n", xml);//the correct xhtmlite version
	strrepl(xml, "<br>", "\n", xml); //html version
	strrepl(xml, "<br></br>", "\n", xml); //incorrect xhtmlite
	strrepl(xml, "<br/>", "\n", xml);//incorrect xhtmlite
	
	strrepl(xml, "<BR />", "\n", xml);//the correct xhtmlite version
	strrepl(xml, "<BR>", "\n", xml); //html version
	strrepl(xml, "<BR></BR>", "\n", xml); //incorrect xhtmlite
	strrepl(xml, "<BR/>", "\n", xml);//incorrect xhtmlite

	gchar *formatted_xml;
	if(markup)
	{
		formatted_xml = g_strdup_printf("<html>%s</html>", xml);
	}
	else
	{
		formatted_xml = g_strdup_printf("%s", xml);
	}
	
	printf("fixed whitespace xml:%s\n", formatted_xml);
	
	offset = gtk_text_iter_get_offset(start);
	xmlTextReaderPtr reader;
	int ret;
	gtk_text_buffer_begin_user_action(textbuffer);
	reader = xmlReaderForMemory((char *)formatted_xml, g_str_get_length((char *)formatted_xml), NULL, "", 0);
	if (reader != NULL) {
		ret = xmlTextReaderRead(reader);
		while (ret == 1) {
			xhtmlite_process_node(textbuffer, reader);
			ret = xmlTextReaderRead(reader);
		}
		xmlFreeTextReader(reader);
		if (ret != 0) {
			printf("Failed to parse xml\n");
		}
	} else {
		printf("Unable to open xml\n");
	}
	gtk_text_buffer_end_user_action(textbuffer);
	printf("\nNow finished reading...\n");	
	return offset;
}

void xhtmlite_import (GtkTextBuffer *textbuffer, gchar *xml, gboolean markup)
{
	GtkTextIter start, end;
	gtk_text_buffer_get_bounds (textbuffer, &start, &end);
	gtk_text_buffer_delete(textbuffer, &start, &end);

	xhtmlite_insert (textbuffer, xml, &start, markup);
	printf("clear undos NOW:)\n");
	xhtmlite_undo_clear(textbuffer);
}

void xhtmlite_process_node(GtkTextBuffer *textbuffer, xmlTextReaderPtr reader)
{
	const xmlChar *name, *value;

	name = (const xmlChar*)g_utf8_strdown((const gchar*)xmlTextReaderConstName(reader), g_str_get_length((const gchar*)xmlTextReaderConstName(reader), -1));
	if (name == NULL)
		name = BAD_CAST "--";

	value = xmlTextReaderConstValue(reader);

	printf("%d %d %s %d %d",
	       xmlTextReaderDepth(reader),
	       xmlTextReaderNodeType(reader),
	       name,
	       xmlTextReaderIsEmptyElement(reader),
	       xmlTextReaderHasValue(reader));
	if (value == NULL)
		printf("\n");
	else {
		if (xmlStrlen(value) > 40)
			printf(" %.40s...\n", value);
		else
			printf(" %s\n", value);
	}

	int nodetype =  xmlTextReaderNodeType(reader);
	switch (nodetype)
	{
		case ((int)XML_READER_TYPE_ELEMENT):
			printf("type_element:%s\n", name);
			if (!xmlStrcasecmp(name, (const xmlChar*)"html"))
			{
				break;
			}
			if (!xmlStrcasecmp(name, (const xmlChar*)"img"))
			{
				printf("img tag found, offset:%d\n", GINT_TO_POINTER(offset));
				const gchar *attribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"src");

				//get pixbuf
				const gchar *classattribute = (const gchar*)xmlTextReaderGetAttribute(reader, (const xmlChar*)"class");
			
				//Set iter
				GtkTextIter image_iter;
				gtk_text_buffer_get_iter_at_offset(textbuffer, &image_iter, offset);
				GdkPixbuf * mypixbuf = xhtmlite_file_decode_pixbuf_from_xhtml(attribute, classattribute);
				if (mypixbuf != NULL)
				{
					gtk_text_buffer_insert_pixbuf(textbuffer, &image_iter, mypixbuf);
				}
				offset += 1;
				break;
			}
			import_tag_stack = g_list_prepend (import_tag_stack, GINT_TO_POINTER(offset)); //push iter on to the list the beginning of the list
			printf("added iter for tag:%s\n", name);
			break;
		case ((int)XML_READER_TYPE_WHITESPACE):
			printf("type_whitespace:%s\n", name);
			break;
		case ((int)XML_READER_TYPE_TEXT):
			printf("type_text:%s\n", name);
			GtkTextIter insert_at;
			gtk_text_buffer_get_iter_at_offset(textbuffer, &insert_at, offset);
			gtk_text_buffer_insert(textbuffer, &insert_at, (const gchar*) value,-1);
			//offset += xmlStrlen(value);
			//FIX: the following lines fix errors where two utf-8 characters only makeup one character.
			//This may need to be changed if you insert into the buffer rather than deleting the buffer and writing the buffer.
			offset += g_str_get_length(value, -1);
			break;
		case((int)XML_READER_TYPE_ATTRIBUTE):
			break;
		case ((int)XML_READER_TYPE_END_ELEMENT):
			printf("type_end_element:%s\n", name);
			if (!xmlStrcasecmp(name, (const xmlChar*)"html"))
			{
				break;
			}
			int *offset_start = g_list_nth_data(import_tag_stack, 0); //pop off offset
			GtkTextIter begin_iter, end_iter;
			gtk_text_buffer_get_iter_at_offset(textbuffer, &begin_iter, GPOINTER_TO_INT(offset_start));
			gtk_text_buffer_get_iter_at_offset(textbuffer, &end_iter, offset);
			import_tag_stack = g_list_remove(import_tag_stack, offset_start); //remove offset

			GtkTextTag *tag;
			GtkTextTagTable *tagtable = gtk_text_buffer_get_tag_table(textbuffer);
			tag = gtk_text_tag_table_lookup(tagtable, (const gchar*) name);
			if(tag != NULL)
			{
				gtk_text_buffer_apply_tag(textbuffer, tag, &begin_iter, &end_iter);
			}
			if(xmlTextReaderHasAttributes(reader))
			{
				//font tag
				if(xmlStrstr(name, (xmlChar *)"font")!= NULL)
				{
					xhtmlite_process_font_tag(textbuffer, reader, &begin_iter, &end_iter);
				}
				//href link
				xhtmlite_process_a_link_tag(textbuffer, reader, &begin_iter, &end_iter);
				//CSS Style tags
				xhtmlite_parse_and_process_style_attribute(textbuffer, reader, &begin_iter, &end_iter);
			}
			break;
		default:
			printf("other");
	}
}

int strrepl(char *InputString,char *StringToFind, char *StringToReplace,char *output)
{
	char *offset = NULL, *CurrentPointer = NULL;
	int insertlen;
	int findlen = g_str_get_length(StringToFind);
	int result = 0;

	if (StringToReplace)
		insertlen = g_str_get_length(StringToReplace);
	else
		insertlen = 0;
	if (output) {
		if (output != InputString)
			memmove(output,InputString,g_str_get_length(InputString)+1);
		InputString = output;
	}
	else
		result = g_str_get_length(InputString)+1;

	while (*InputString)        {
		offset = strstr(!offset?InputString:CurrentPointer,StringToFind);
		if (offset == NULL)
			break;
		CurrentPointer = (offset + (output ? insertlen : findlen));
		if (output) {
			strcpy (offset, (offset + findlen));
			memmove (offset + insertlen,
				 offset, g_str_get_length (offset) + 1);
			if (insertlen)
				memcpy (offset, StringToReplace, insertlen);
			result++;
		}
		else {
			result -= findlen;
			result += insertlen;
		}
	}
	return result;
}

static GtkWidget *xhtmlite_dialog_symbol = NULL;
void xhtmlite_open_symbol_dialog (GtkWindow *parent_window, GtkTextBuffer *textbuffer)
{
	if(xhtmlite_dialog_symbol == NULL)
	{
		xhtmlite_dialog_symbol = create_xhtmlite_dialog_symbol();
		gtk_window_set_transient_for(GTK_WINDOW(xhtmlite_dialog_symbol), parent_window);
		GtkWidget *xhtmlite_symbol_combobox_unicodetype = lookup_widget(GTK_WIDGET(xhtmlite_dialog_symbol), "xhtmlite_symbol_combobox_unicodetype");
		gtk_combo_box_set_active(GTK_COMBO_BOX(xhtmlite_symbol_combobox_unicodetype), 2);
		xhtmlite_symbol_combobox_unicodetype_on_changed(xhtmlite_symbol_combobox_unicodetype, NULL);
	}

	gtk_widget_show(xhtmlite_dialog_symbol);
	GtkWidget *xhtmlite_symbol_label_output = lookup_widget(GTK_WIDGET(xhtmlite_dialog_symbol), "xhtmlite_symbol_label_output");
	gtk_label_set_text(GTK_LABEL(xhtmlite_symbol_label_output), "");
	//gtk_widget_show(xhtml_symbol_table);
	if (gtk_dialog_run (GTK_DIALOG (xhtmlite_dialog_symbol)) == GTK_RESPONSE_OK)
	{
		const gchar *output_text = gtk_label_get_text(GTK_LABEL(xhtmlite_symbol_label_output));
		if(strcmp(output_text, "(none)"))
		{
			g_printf("ADDED:%s\n", output_text);
			gtk_text_buffer_insert_at_cursor(textbuffer, output_text, -1);
		}
	}
	gtk_widget_hide(xhtmlite_dialog_symbol);
}

void xhtmlite_on_button_symbol_activate (GtkButton *button, gpointer user_data)
{
	GtkWidget *xhtmlite_symbol_label_selected = lookup_widget(GTK_WIDGET(button), "xhtmlite_symbol_label_selected");
	const gchar *label = gtk_button_get_label(button);
	gchar buf[40];
	g_sprintf(buf, "<span size=\"36000\">%s</span>\0", label);
	g_printf("ACTIVATED:%s\n", buf);
	gtk_label_set_markup (GTK_LABEL(xhtmlite_symbol_label_selected), buf);
}

void xhtmlite_on_button_symbol_clicked (GtkButton *button, gpointer user_data)
{
	GtkWidget *xhtml_symbol_label_output = lookup_widget(GTK_WIDGET(button), "xhtmlite_symbol_label_output");
	const gchar *label =	gtk_button_get_label(button);
	const gchar *output_text = gtk_label_get_text(GTK_LABEL(xhtml_symbol_label_output));
	gchar buf[48];
	if(!strcmp(output_text, "(none)"))
	{
		g_sprintf(buf, "%s", label);
	}
	else
	{
		g_sprintf(buf, "%s%s", output_text, label);
	}
	g_printf("OUTPUT:%s\n", buf);
	gtk_label_set_text (GTK_LABEL(xhtml_symbol_label_output), buf);
}

static GtkWidget *xhtmlite_dialog_www_link = NULL;
void xhtmlite_insert_link (GtkWindow *parent_window, GtkTextBuffer *textbuffer)
{
	if(xhtmlite_dialog_www_link == NULL)
	{
		xhtmlite_dialog_www_link = create_xhtmlite_dialog_www_link();
		gtk_window_set_transient_for(GTK_WINDOW(xhtmlite_dialog_www_link), parent_window);
	}
	
	GtkWidget *xhtmlite_www_link_entry = lookup_widget(GTK_WIDGET(xhtmlite_dialog_www_link), "xhtmlite_www_link_entry");
	GtkTextIter beginselect, endselect;
	gtk_text_buffer_get_selection_bounds(textbuffer, &beginselect, &endselect);	

	/*get old tag on first iter*/
	GSList *tags = gtk_text_iter_get_tags(&beginselect);
	guint len = g_slist_length(tags);
	int n =0;
	for(n = 0; n < len; n++)
	{
		GtkTextTag *tag = g_slist_nth_data (tags, n);
		gchar *old_link = g_object_get_data(G_OBJECT(tag), "link");
		if(g_str_get_length(old_link) > 0)
		{
			gtk_entry_set_text(GTK_ENTRY(xhtmlite_www_link_entry), old_link);
		}
	}
	
	//set entry if there is already a link
	if (gtk_dialog_run (GTK_DIALOG (xhtmlite_dialog_www_link)) == GTK_RESPONSE_OK)
	{
		const gchar *link = gtk_entry_get_text(GTK_ENTRY(xhtmlite_www_link_entry));
		xhtmlite_add_www_link(textbuffer, (gchar *)link, &beginselect, &endselect);
	}
	gtk_widget_hide(xhtmlite_dialog_www_link);
}
